/*(c) 1997 N Grassly and A Rambaut
 *	Department of Zoology
 *	Oxford University
 *	pseq-gen.c
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>

#ifdef __MWERKS__
#include <console.h>
#endif

#include "global.h"
#include "treefile.h"
#include "pevolve.h"
#include "pmodels.h"
#include "progress.h"
#include "random.h"

#define PROGRAM_NAME "pseq-gen"
#define VERSION_NUMBER "Version 1.0"

char *ModelNames[4]={
	"PAM",
	"JTT",
	"MTREV",
	"GENERAL"
};

int printHelp, treeFile, numDatasets, numTrees;
int scaleTrees, scaleBranches, mfreqSet;
double treeScale, branchScale;
char treeFileName[256];

/* prototypes */

static void PrintTitle();
static void PrintUsage();
static void PrintVerbose();
static void ReadParams();

/* functions */

static void PrintTitle()
{
	fprintf(stderr, "Protein Sequence Generator - %s\n", PROGRAM_NAME);
	fprintf(stderr, "%s\n", VERSION_NUMBER);
	fprintf(stderr, "(c) Copyright, 1997 Nick Grassly and Andrew Rambaut\n");
	fprintf(stderr, "Department of Zoology, University of Oxford\n");
	fprintf(stderr, "South Parks Road, Oxford OX1 3PS, U.K.\n\n");
}

static void PrintUsage()
{ 
	fprintf(stderr, "Usage: seq-gen [-m MODEL] [-l #] [-n #] [-s # | -d #] [-g #]\n               [-f #ala,arg...] [-p] [-q] [-h] [treefile]\n");
	fprintf(stderr, "  -m: MODEL = PAM, JTT, MTREV [default = PAM]\n");
	fprintf(stderr, "  -l: # = sequence length [default = 1000].\n");
	fprintf(stderr, "  -n: # = simulated datasets per tree [default = 1].\n");
	fprintf(stderr, "  -s: # = branch length scaling factor [default = 1.0].\n");
	fprintf(stderr, "  -d: # = total tree scale [default = use branch lengths].\n");
	fprintf(stderr, "  -g: # = shape for gamma rate heterogeneity [default = none].\n");
	fprintf(stderr, "  -f: # = amino acid frequencies e=equal [default = matrix freqs].\n");
	fprintf(stderr, "  -p: Use the relaxed PHYLIP format\n");
	fprintf(stderr, "  -h: Give this help message\n");
	fprintf(stderr, "  -q: Quiet\n");
	fprintf(stderr, "  -v: Verbose\n");
	fprintf(stderr, "  treefile: name of tree file [default = trees on stdin]\n\n");
}


void ReadParams(int argc, char **argv)
{
	int i, j;
	char ch, *P, st[3];
	
	model=PAM;

	scaleTrees=0;
	treeScale=0.0;
	scaleBranches=0;
	branchScale=0.0;
	mfreqSet=1;
	rateHetero=NoRates;
	gammaShape=1.0;
	
	
	numBases=1000;
	numDatasets=1;
	
 	verbose=0;
	relaxedPhylip=0;
	quiet=0;

	treeFile=0;
		
	for (i=1; i<argc; i++) {
		P=argv[i];
		if (*P!='-') {
			if (treeFile) {
				fprintf(stderr, "Illegal command parameter: %s\n\n", argv[i]);
				PrintUsage();
				exit(0);
			}
			treeFile=1;
			strcpy(treeFileName, argv[i]);
		} else {
			while (*P) {
				*P=toupper(*P);
				P++;
			}
			P=argv[i];
			P++;
			ch=*P;
			P++;
			switch (ch) {
				case 'H':
					printHelp=1;
				break;
				case 'M':
					if (GetStrParam(argc, argv, &i, P, st, 3)) {
						fprintf(stderr, "Bad (or missing) Model Code: %s\n\n", argv[i]);
						exit(0);
					}
					
					P=st;
					model=-1;
					for (j=PAM; j<GENERAL; j++) {
						if (strncmp(P, ModelNames[j], 3)==0)
							model=j;
					}
					if (model==-1) {
						fprintf(stderr, "Unknown Model: %s\n\n", argv[i]);
						exit(0);
					}

				break;
				case 'L':
					if (GetIntParams(argc, argv, &i, P, 1, &numBases) || numBases<1) {
						fprintf(stderr, "Bad (or missing) sequence length: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'N':
					if (GetIntParams(argc, argv, &i, P, 1, &numDatasets) || numDatasets<1) {
						fprintf(stderr, "Bad (or missing) number of datasets: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'G':
					rateHetero=GammaRates;
					if (GetDoubleParams(argc, argv, &i, P, 1, &gammaShape)) {
						fprintf(stderr, "Bad Gamma Shape: %s\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'F':
					mfreqSet=0;
					if (*P=='E'){
						for(i=0;i<NUM_AA;i++)
							freq[i]=0.05;
					}else{
						if (GetDoubleParams(argc, argv, &i, P, NUM_AA, freq)) {
							fprintf(stderr, "Bad Amino Acid Frequencies: %s\n\n", argv[i]);
							exit(0);
						}
					}
				break;
				case 'D':
					scaleTrees=1;
					if (GetDoubleParams(argc, argv, &i, P, 1, &treeScale) || treeScale<=0.0) {
						fprintf(stderr, "Bad Total Tree Scale: %s\n\n", argv[i]);
						exit(0);
					}
					if (scaleBranches) {
						fprintf(stderr, "You can't specify both the -d and -s options\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'S':
					scaleBranches=1;
					if (GetDoubleParams(argc, argv, &i, P, 1, &branchScale) || branchScale<=0.0) {
						fprintf(stderr, "Bad Branch Length Scale: %s\n\n", argv[i]);
						exit(0);
					}
					if (scaleTrees) {
						fprintf(stderr, "You can't specify both the -d and -s options\n\n", argv[i]);
						exit(0);
					}
				break;
				case 'P':
					relaxedPhylip=1;
				break;
				case 'Q':
					quiet=1;
				break;
				case 'V':
					verbose=1;
				break;
				default:
					fprintf(stderr, "Illegal command parameter: %s\n\n", argv[i]);
					PrintUsage();
					exit(0);
				break;
			}
		}
	}
}

void PrintVerbose()
{
int i;
	fprintf(stderr, "Simulating %d taxa, %d bases\n", numTaxa, numBases);
	fprintf(stderr, "  for %d tree(s) with %d dataset(s) per tree\n\n", numTrees, numDatasets);
	
	if (scaleTrees) {
		fprintf(stderr, "Branch lengths of trees scaled so that tree is %G from root to tip\n\n", treeScale);
	} else if (scaleBranches) {
		fprintf(stderr, "Branch lengths of trees multiplied by %G\n\n", branchScale);
	} else {
		fprintf(stderr, "Branch lengths assumed to be number of substitutions per site\n");
	}
	fprintf(stderr, "Substitution model is %s\n\n", ModelNames[model]);
	if (rateHetero==GammaRates) {
		fprintf(stderr, "Discrete gamma rate heterogeneity:\n");
		fprintf(stderr, "    shape = %f\n", gammaShape);
	} else
		fprintf(stderr, "Rate homogeneity of sites.\n");
	fprintf(stderr, "Amino Acid Frequencies\n");
	for(i=0;i<NUM_AA;i++)
		fprintf(stderr, "%c: %f\t", aminoAcids[i], freq[i]);
	fprintf(stderr, "\n\n");
	fflush(stderr);
	if(verbose){
		printEigen();
		printPij();
		fflush(stderr);
	}
}

void main(int argc, char **argv)
{
	int i, treeNo;
	FILE *tree_fv;
	TTree *tree;
	clock_t totalStart;
	double totalSecs, scale;
	
#ifdef __MWERKS__
	argc=ccommand(&argv);
#endif

	if (setvbuf( stderr, NULL, _IOLBF, 512 ) ||
		setvbuf( stdout, NULL, _IOLBF, 512 ) ) {
		fprintf(stderr, "Failed to buffer stdout and stderr\n");
		exit(0);
	}
 	
	SetSeed(time(NULL));
	totalStart = clock();

	ReadParams(argc, argv);

	if (printHelp) {
 		PrintTitle();
		PrintUsage();
		exit(0);
	}

	if (!quiet)
 		PrintTitle();

	SetModel(model);
	
	numTaxa=-1;
	scale=1000.0;/* therefore bl=1 => 1 change in hundred */
	
	if ((tree=NewTree())==NULL) {
		fprintf(stderr, "Out of memory\n");
		exit(0);
	}
	
	if (treeFile) {
		if ( (tree_fv=fopen(treeFileName, "rt"))==NULL ) {
			fprintf(stderr, "Error opening tree file: '%s'\n", treeFileName);
			exit(0);
		}
	} else 
		tree_fv=stdin;
		
	numTrees=CountTrees(tree_fv);
	
	treeNo=0;
	do {
		ReadTree(tree_fv, tree, treeNo+1);
		
		if (treeNo==0) {
			numTaxa=tree->numTips;
			CreateSequences(tree);
			
			if (scaleTrees) { 
				if (!tree->rooted) {
					fprintf(stderr, "To scale tree length, they must be rooted and ultrametric.\n");
					exit(0);
				}
				scale*=(treeScale/tree->totalLength);
			} else if (scaleBranches)
				scale*=branchScale;

			if (!quiet) {
				PrintVerbose();
				InitProgressBar(numTrees*numDatasets);
				DrawProgressBar();
			}
		} else if (tree->numTips!=numTaxa) {
			fprintf(stderr, "All trees must have the same number of tips.\n");
			exit(0);
		}
				
		for (i=0; i<numDatasets; i++) {
			SetCategories();
			EvolveSequences(tree, scale);
			WriteSequences(stdout, tree);
			if (!quiet)
				ProgressBar();
		}
				
		DisposeTree(tree);
		treeNo++;
	} while (IsTreeAvail(tree_fv));
	
	FreeTree(tree);
	
	totalSecs = (double)(clock() - totalStart) / CLOCKS_PER_SEC;
	if (!quiet) {
		fprintf(stderr, "Time taken: %G seconds\n", totalSecs);
		if (verboseMemory)
			fprintf(stderr, "Total memory used: %ld\n", totalMem);
	}
}
